__all__ = ['Process', 'Interaction', 'GeneticInteraction',
           'MolecularInteraction', 'TemplateReaction',
           'Control', 'Conversion', 'Catalysis',
           'TemplateReactionRegulation', 'Modulation',
           'ComplexAssembly', 'BiochemicalReaction',
           'Degradation', 'Transport', 'TransportWithBiochemicalReaction']

from .base import Entity


class Process(Entity):
    """BioPAX Process."""
    def __init__(self,
                 controlled_of=None,
                 step_process_of=None,
                 pathway_component_of=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.controlled_of = controlled_of
        self.step_process_of = step_process_of
        self.pathway_component_of = pathway_component_of


class Interaction(Process):
    """BioPAX Interaction."""
    def __init__(self,
                 participant=None,
                 interaction_type=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.participant = participant
        self.interaction_type = interaction_type


class GeneticInteraction(Interaction):
    """BioPAX GeneticInteraction."""
    pass


class MolecularInteraction(Interaction):
    """BioPAX MolecularInteraction."""
    pass


class TemplateReaction(Interaction):
    """BioPAX TemplateReaction."""
    list_types = Interaction.list_types + ['product']

    def __init__(self,
                 template=None,
                 product=None,
                 template_direction=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.template = template
        self.product = product
        self.template_direction = template_direction


class Control(Interaction):
    """BioPAX Control."""
    list_types = Interaction.list_types + ['controller']

    def __init__(self,
                 control_type=None,
                 controller=None,
                 controlled=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.control_type = control_type
        self.controller = controller
        self.controlled = controlled


class Conversion(Interaction):
    """BioPAX Conversion."""
    list_types = Interaction.list_types + \
        ['left', 'right', 'participant_stoichiometry']

    def __init__(self,
                 left=None,
                 right=None,
                 conversion_direction=None,
                 participant_stoichiometry=None,
                 spontaneous=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.left = left
        self.right = right
        self.conversion_direction = conversion_direction
        self.participant_stoichiometry = participant_stoichiometry
        self.spontaneous = spontaneous


class Catalysis(Control):
    """BioPAX Catalysis."""
    def __init__(self,
                 catalysis_direction=None,
                 cofactor=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.catalysis_direction = catalysis_direction
        self.cofactor = cofactor


class TemplateReactionRegulation(Control):
    """BioPAX TemplateReactionRegulation."""
    pass


class Modulation(Control):
    """BioPAX Modulation."""
    pass


class ComplexAssembly(Conversion):
    """BioPAX ComplexAssembly."""
    pass


class BiochemicalReaction(Conversion):
    """BioPAX BiochemicalReaction."""
    def __init__(self,
                 delta_s=None,
                 delta_h=None,
                 delta_g=None,
                 k_e_q=None,
                 e_c_number=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.delta_s = delta_s
        self.delta_h = delta_h
        self.delta_g = delta_g
        self.k_e_q = k_e_q
        self.e_c_number = e_c_number


class Degradation(Conversion):
    """BioPAX Degradation."""
    pass


class Transport(Conversion):
    """BioPAX Transport."""
    pass


class TransportWithBiochemicalReaction(BiochemicalReaction):
    """BioPAX TransportWithBiochemicalReaction."""
    pass




__all__ = ['BioPaxObject', 'Entity', 'Pathway', 'Gene', 'Unresolved']

from ..xml_util import *


class Unresolved:
    def __init__(self, obj_id):
        self.obj_id = obj_id


class BioPaxObject:
    """Generic BioPAX Object. It is the parent class of all more specific
    BioPAX classes."""
    list_types = ['xref', 'comment']
    xml_types = {}

    def __init__(self, uid, name=None, comment=None, xref=None):
        self.uid = uid
        # TODO: is name in the right place here?
        self.name = name
        self.comment = comment
        # TODO: is xref in the right place here?
        self.xref = xref

    @classmethod
    def from_xml(cls, element):
        uid = get_id_or_about(element)
        kwargs = {'uid': uid}
        for key in cls.list_types:
            kwargs[key] = []
        for child in element.getchildren():
            key = get_attr_tag(child)
            if is_datatype(child.attrib, nssuffix('xsd', 'string')) \
                    or is_datatype(child.attrib, nssuffix('xsd', 'int')) \
                    or is_datatype(child.attrib, nssuffix('xsd', 'float')):
                val_to_add = child.text
            else:
                res = get_resource(child.attrib)
                val_to_add = Unresolved(res)
            if key in cls.list_types:
                kwargs[key].append(val_to_add)
            else:
                kwargs[key] = val_to_add
        return cls(**kwargs)

    def to_xml(self):
        id_type = 'about' if is_url(self.uid) else 'ID'
        element = makers['bp'](self.__class__.__name__,
                               **{nselem('rdf', id_type): self.uid})
        for attr in [a for a in dir(self)
                     if not a.startswith('_')
                     and a not in {'list_types', 'xml_types',
                                   'to_xml', 'from_xml', 'uid'}]:
            val = getattr(self, attr)
            if val is None:
                continue
            if isinstance(val, list):
                for v in val:
                    child_elem = self._simple_to_xml(attr, v)
                    if child_elem is not None:
                        element.append(child_elem)
            else:
                child_elem = self._simple_to_xml(attr, val)
                if child_elem is not None:
                    element.append(child_elem)
        return element

    def _simple_to_xml(self, attr, val):
        if isinstance(val, BioPaxObject):
            child_elem = makers['bp'](
                snake_to_camel(attr),
                **{nselem('rdf', 'resource'):
                    ('#%s' % val.uid) if not is_url(val.uid) else val.uid}
            )
            return child_elem
        elif isinstance(val, str):
            xml_type = self.xml_types.get(attr, 'string')
            child_elem = makers['bp'](
                snake_to_camel(attr),
                val,
                **{nselem('rdf', 'datatype'): nssuffix('xsd', xml_type)}
            )
            return child_elem
        return None


class Entity(BioPaxObject):
    """BioPAX Entity."""
    list_types = BioPaxObject.list_types + \
        ['evidence', 'data_source']

    def __init__(self,
                 standard_name=None,
                 display_name=None,
                 all_names=None,
                 participant_of=None,
                 availability=None,
                 data_source=None,
                 evidence=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.standard_name = standard_name
        self.display_name = display_name
        self.all_names = all_names
        self.participant_of = participant_of
        self.availability = availability
        self.data_source = data_source
        self.evidence = evidence


class Gene(Entity):
    """BioPAX Gene"""
    def __init__(self, organism, **kwargs):
        super().__init__(**kwargs)
        self.organism = organism


class Pathway(Entity):
    """BioPAX Pathway."""
    list_types = Entity.list_types + ['pathway_component']

    def __init__(self,
                 pathway_component=None,
                 pathway_order=None,
                 organism=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.pathway_component = pathway_component
        self.pathway_order = pathway_order
        self.organism = organism




import os
from pybiopax.biopax import *
from pybiopax import model_from_owl_file

here = os.path.dirname(os.path.abspath(__file__))
test_file = os.path.join(here, 'biopax_test.owl')


def test_process_owl():
    model = model_from_owl_file(test_file)
    assert len(model.objects) == 58027, len(model.objects)
    assert 'BiochemicalReaction_a75d6aaebf5be718d981d95355ced14a' in \
        model.objects
    bcr = model.objects['BiochemicalReaction_a75d6aaebf5be718d981d95355ced14a']
    assert bcr.display_name == \
        'STATE_TRANSITION_LEFT__Raf-RIGHT__Raf-_r21'
    assert isinstance(bcr.left, list)
    assert isinstance(bcr.left[0], Protein)
    assert isinstance(bcr.right, list)
    assert isinstance(bcr.right[0], Protein)

    pub = model.objects['PublicationXref_http___www_phosphosite_org_'
                        'phosphosite_owl_recid_11958410']
    assert len(pub.comment) == 1
    assert pub.comment[0].startswith('REPLACED')
    assert pub.url.startswith('http://www.phosphosite')
    assert pub.year == '2010', pub.year



__all__ = ['BioPaxModel']

import tqdm
from . import *
from ..xml_util import has_ns, get_id_or_about, get_tag, wrap_xml_elements


class BioPaxModel:
    """BioPAX Model.

    Parameters
    ----------
    objects : dict
        A dict of BioPaxObject instances keyed by their URI string.

    Attributes
    ----------
    objects : dict
        A dict of BioPaxObject instances keyed by their URI string
        that are part of the model.
    """
    def __init__(self, objects):
        self.objects = objects

    @classmethod
    def from_xml(cls, tree):
        """Return a BioPAX Model from an OWL/XML element tree."""
        objects = {}
        for element in tqdm.tqdm(tree.getchildren(),
                                 desc='Processing OWL elements'):
            if not has_ns(element, 'bp'):
                continue
            id = get_id_or_about(element)
            obj_cls = globals()[get_tag(element)]
            obj = obj_cls.from_xml(element)
            objects[id] = obj

        for obj_id, obj in objects.items():
            for attr in [a for a in dir(obj) if not a.startswith('__')]:
                val = getattr(obj, attr)
                resolved_val = resolve_value(objects, val)
                setattr(obj, attr, resolved_val)

        return cls(objects)

    def to_xml(self):
        """Return an OWL string from the content of the model."""
        elements = [obj.to_xml() for obj in self.objects.values()]
        return wrap_xml_elements(elements)

    def get_objects_by_type(self, obj_type):
        for obj in self.objects.values():
            if isinstance(obj, obj_type):
                yield obj


def resolve_value(objects, val):
    if isinstance(val, Unresolved):
        if val.obj_id not in objects:
            resolved_val = val.obj_id
        else:
            resolved_val = objects[val.obj_id]
    elif isinstance(val, list):
        resolved_val = [resolve_value(objects, v) for v in val]
    else:
        resolved_val = val
    return resolved_val



__all__ = ['UtilityClass', 'Evidence', 'Provenance',
           'EntityFeature', 'ModificationFeature', 'FragmentFeature',
           'BindingFeature', 'BioSource',
           'ExperimentalForm', 'SequenceLocation', 'SequenceInterval',
           'SequenceSite', 'PathwayStep', 'BiochemicalPathwayStep',
           'Xref', 'PublicationXref',
           'UnificationXref', 'RelationshipXref', 'EntityReference',
           'ProteinReference', 'RnaReference', 'DnaReference',
           'SmallMoleculeReference', 'RnaRegionReference',
           'DnaRegionReference', 'SequenceEntityReference',
           'ChemicalStructure', 'ChemicalConstant', 'DeltaG', 'KPrime',
           'Stoichiometry', 'ControlledVocabulary',
           'CellularLocationVocabulary', 'EntityReferenceTypeVocabulary',
           'EvidenceCodeVocabulary', 'ExperimentalFormVocabulary',
           'InteractionVocabulary', 'PhenotypeVocabulary',
           'RelationshipTypeVocabulary', 'SequenceModificationVocabulary',
           'SequenceRegionVocabulary', 'TissueVocabulary', 'CellVocabulary',
           'Score']

from .base import BioPaxObject


class UtilityClass(BioPaxObject):
    """BioPAX UtilityClass."""
    def __int__(self, **kwargs):
        super().__init__(**kwargs)


class Evidence(UtilityClass):
    """BioPAX Evidence."""
    def __init__(self,
                 confidence=None,
                 evidence_code=None,
                 experimental_form=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.confidence = confidence
        self.evidence_code = evidence_code
        self.experimental_form = experimental_form


class Provenance(UtilityClass):
    """BioPAX Provenance."""
    def __init__(self,
                 standard_name=None,
                 display_name=None,
                 all_names=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.standard_name = standard_name
        self.display_name = display_name
        self.all_names = all_names


class EntityFeature(UtilityClass):
    """BioPAX UtilityClass."""
    list_types = UtilityClass.list_types + ['evidence']

    def __init__(self,
                 evidence=None,
                 owner_entity_reference=None,
                 feature_of=None,
                 not_feature_of=None,
                 feature_location=None,
                 member_feature=None,
                 feature_location_type=None,
                 member_feature_of=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.evidence = evidence
        self.owner_entity_reference = owner_entity_reference
        self.feature_of = feature_of
        self.not_feature_of = not_feature_of
        self.feature_location = feature_location
        self.member_feature = member_feature
        self.feature_location_type = feature_location_type
        self.member_feature_of = member_feature_of


class ModificationFeature(EntityFeature):
    """BioPAX ModificationFeature."""
    def __init__(self,
                 modification_type=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.modification_type = modification_type

    def __str__(self):
        return '%s(%s%s)' % (self.__class__.__name__,
                             self.modification_type if self.modification_type
                             else '',
                             '@%s' % self.feature_location
                             if self.feature_location else '')

    def __repr__(self):
        return str(self)


class FragmentFeature(EntityFeature):
    """BioPAX FragmentFeature."""
    pass


class BindingFeature(EntityFeature):
    """BioPAX BindingFeature."""
    def __init__(self,
                 binds_to=None,
                 intra_molecular=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.binds_to = binds_to
        self.intra_molecular = intra_molecular


class ChemicalConstant(UtilityClass):
    xml_types = {'ionic_strength': 'float',
                 'ph': 'float',
                 'p_mg': 'float',
                 'temperature': 'float'}

    def __init__(self,
                 ionic_strength=None,
                 ph=None,
                 p_mg=None,
                 tempterature=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.ionic_strength = ionic_strength
        self.ph = ph
        self.p_mg = p_mg
        self.temperature = tempterature


class DeltaG(ChemicalConstant):
    xml_types = {'delta_g_prime0': 'float'}

    def __init__(self,
                 delta_g_prime0=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.delta_g_prime0 = delta_g_prime0


class KPrime(ChemicalConstant):
    """BioPAX KPrime."""
    def __init__(self, k_prime, **kwargs):
        super().__init__(**kwargs)
        self.k_prime = k_prime


class ChemicalStructure(UtilityClass):
    def __init__(self,
                 structure_format=None,
                 structure_data=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.structure_format = structure_format
        self.structure_data = structure_data


class BioSource(UtilityClass):
    """BioPAX BioSource."""
    def __init__(self,
                 cell_type=None,
                 tissue=None,
                 standard_name=None,
                 display_name=None,
                 all_names=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.cell_type = cell_type
        self.tissue = tissue
        self.standard_name = standard_name
        self.display_name = display_name
        self.all_names = all_names


class ExperimentalForm(UtilityClass):
    """BioPAX ExperimentalForm."""
    def __init__(self,
                 experimental_form_entity=None,
                 experimental_form_description=None,
                 experimental_feature=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.experimental_form_entity = experimental_form_entity
        self.experimental_form_description = experimental_form_description
        self.experimental_feature = experimental_feature


class SequenceLocation(UtilityClass):
    """BioPAX SequenceLocation."""
    def __init__(self,
                 region_type=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.region_type = region_type


class SequenceInterval(SequenceLocation):
    """BioPAX SequenceInterval."""
    def __init__(self,
                 sequence_interval_begin=None,
                 sequence_interval_end=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.sequence_interval_begin = sequence_interval_begin
        self.sequence_interval_end = sequence_interval_end

    def __str__(self):
        return '%s(%s-%s)' % (self.__class__.__name__,
                              self.sequence_interval_begin,
                              self.sequence_interval_end)

    def __repr__(self):
        return str(self)


class SequenceSite(SequenceLocation):
    """BioPAX SequenceSite."""
    xml_types = {'sequence_position': 'int'}

    def __init__(self,
                 position_status=None,
                 sequence_position=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.position_status = position_status
        self.sequence_position = sequence_position

    def __str__(self):
        return '%s(%s)' % (self.__class__.__name__,
                           self.sequence_position)

    def __repr__(self):
        return str(self)


class PathwayStep(UtilityClass):
    """BioPAX PathwayStep."""
    list_types = UtilityClass.list_types + ['evidence']

    def __init__(self,
                 step_process=None,
                 next_step=None,
                 next_step_of=None,
                 pathway_order_of=None,
                 evidence=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.step_process = step_process
        self.next_step = next_step
        self.next_step_of = next_step_of
        self.pathway_order_of = pathway_order_of
        self.evidence = evidence


class BiochemicalPathwayStep(PathwayStep):
    """BioPAX BiochemicalPathwayStep."""
    def __init__(self,
                 step_conversion=None,
                 step_direction=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.step_conversion = step_conversion
        self.step_direction = step_direction


class Xref(UtilityClass):
    """BioPAX Xref."""
    def __init__(self,
                 db=None,
                 id=None,
                 db_version=None,
                 id_version=None,
                 xref_of=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.db = db
        self.db_version = db_version
        self.id_version = id_version
        self.id = id
        self.xref_of = xref_of


class PublicationXref(Xref):
    """BioPAX PublicationXref."""
    xml_types = {'year': 'int'}

    def __init__(self,
                 title=None,
                 url=None,
                 source=None,
                 author=None,
                 year=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.title = title
        self.url = url
        self.source = source
        self.author = author
        self.year = year


class UnificationXref(Xref):
    """BioPAX UnificationXref."""
    pass


class RelationshipXref(Xref):
    """BioPAX RelationshipXref."""
    def __init__(self,
                 relationship_type=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.relationship_type = relationship_type


class Score(UtilityClass):
    """BioPAX Score."""
    def __init__(self,
                 score_source=None,
                 value=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.score_source = score_source
        self.value = value


class EntityReference(UtilityClass):
    """BioPAX EntityReference."""
    list_types = UtilityClass.list_types + \
        ['evidence', 'entity_feature', 'member_entity_reference']

    def __init__(self,
                 entity_feature=None,
                 entity_reference_of=None,
                 evidence=None,
                 entity_reference_type=None,
                 member_entity_reference=None,
                 owner_entity_reference=None,
                 xref=None,
                 standard_name=None,
                 display_name=None,
                 all_names=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.entity_feature = entity_feature
        self.entity_reference_of = entity_reference_of
        self.evidence = evidence
        self.entity_reference_type = entity_reference_type
        self.member_entity_reference = member_entity_reference
        self.owner_entity_reference = owner_entity_reference
        # TODO: is xref in the right location here?
        self.xref = xref
        self.standard_name = standard_name
        self.display_name = display_name
        self.all_names = all_names


class SequenceEntityReference(EntityReference):
    """BioPAX SequenceEntityReference."""
    def __init__(self,
                 organism=None,
                 sequence=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.organism = organism
        self.sequence = sequence


class RnaReference(SequenceEntityReference):
    """BioPAX RnaReference."""
    pass


class RnaRegionReference(EntityReference):
    """BioPAX RnaRegionReference."""
    pass


class ProteinReference(SequenceEntityReference):
    """BioPAX ProteinReference."""
    pass


class SmallMoleculeReference(EntityReference):
    """BioPAX SmallMoleculeReference."""
    xml_types = {'molecular_weight': 'float'}

    def __init__(self,
                 structure=None,
                 chemical_formula=None,
                 molecular_weight=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.structure = structure
        self.chemical_formula = chemical_formula
        self.molecular_weight = molecular_weight


class DnaReference(SequenceEntityReference):
    """BioPAX DnaReference."""
    pass


class DnaRegionReference(EntityReference):
    """BioPAX DnaRegionReference."""
    pass


class ChemicalStructure(UtilityClass):
    """BioPAX ChemicalStructure."""
    def __init__(self,
                 structure_format=None,
                 structure_data=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.structure_format = structure_format
        self.structure_data = structure_data


class Stoichiometry(UtilityClass):
    """BioPAX Stoichiometry."""
    xml_types = {'stoichiometric_coefficient': 'float'}

    def __init__(self,
                 stoichiometric_coefficient=None,
                 physical_entity=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.stoichiometric_coefficient = stoichiometric_coefficient
        self.physical_entity = physical_entity


class ControlledVocabulary(UtilityClass):
    """BioPAX ControlledVocabulary."""
    list_types = UtilityClass.list_types + ['term']

    def __init__(self, term=None, **kwargs):
        super().__init__(**kwargs)
        self.term = term

    def __str__(self):
        if self.term:
            terms_str = ', '.join(['"%s"' % t for t in self.term])
        else:
            terms_str = ''
        return '%s(%s)' % (self.__class__.__name__, terms_str)

    def __repr__(self):
        return str(self)


class ExperimentalFormVocabulary(ControlledVocabulary):
    """BioPAX ExperimentalFormVocabulary."""
    pass


class RelationshipTypeVocabulary(ControlledVocabulary):
    """BioPAX RelationshipTypeVocabulary."""
    pass


class CellularLocationVocabulary(ControlledVocabulary):
    """BioPAX CellularLocationVocabulary."""
    pass


class EntityReferenceTypeVocabulary(ControlledVocabulary):
    """BioPAX EntityReferenceTypeVocabulary."""
    pass


class PhenotypeVocabulary(ControlledVocabulary):
    """BioPAX PhenotypeVocabulary."""
    pass


class TissueVocabulary(ControlledVocabulary):
    """BioPAX TissueVocabulary."""
    pass


class EvidenceCodeVocabulary(ControlledVocabulary):
    """BioPAX EvidenceCodeVocabulary."""
    pass


class SequenceModificationVocabulary(ControlledVocabulary):
    """BioPAX SequenceModificationVocabulary."""
    pass


class SequenceRegionVocabulary(ControlledVocabulary):
    """BioPAX SequenceRegionVocabulary."""
    pass


class InteractionVocabulary(ControlledVocabulary):
    """BioPAX InteractionVocabulary."""
    pass


class CellVocabulary(ControlledVocabulary):
    """BioPAX CellVocabulary."""
    pass
